###############################################################################
#
# Functions and classes for mathematical mappings.
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

from numbers import Number

###############################################################################


class LinearMap:
    """
    Linear map from input `x` to output `y` with optional boundaries.

    if x < x1 and y1_fill is not None:
        # Left boundary.
        y = y1_fill
    elif x > x2 and y2_fill is not None:
        # Right boundary.
        y = y2_fill
    else:
        # Linear mapping between boundaries.
        # x == x1 maps to y == y1
        # x == x2 maps to y == y2
        a = (y2 - y1) / (x2 - x1)
        b = y1 - a * x1
        y = a * x + b
    """
    def __init__(self, x1, y1, x2, y2, y1_fill=None, y2_fill=None):
        """
        :param x1: Left point for linear mapping (x1, y1)
        :param y1: Left point for linear mapping (x1, y1)
        :param x2: Right point for linear mapping (x2, y2)
        :param y2: Right point for linear mapping (x2, y2)
        :param y1_fill:
            Left boundary for linear mapping: If `x < x1` then `y = y1_fill`.
            If `None` then continue the linear mapping beyond `x1`.
        :param y2_fill:
            Right boundary for linear mapping: If `x > x2` then `y = y2_fill`.
            If `None` then continue the linear mapping beyond `x2`.
        :raises
            `TypeError`: An argument is of the wrong type.
            `ValueError`: An argument has an invalid value.
        """
        # Check argument types.
        number_types = (float, int)
        if not isinstance(x1, number_types):
            raise TypeError('Arg `x1` is not a float or int.')
        if not isinstance(x2, number_types):
            raise TypeError('Arg `x2` is not a float or int.')
        if not isinstance(y1, number_types):
            raise TypeError('Arg `y1` is not a float or int.')
        if not isinstance(y2, number_types):
            raise TypeError('Arg `y2` is not a float or int.')
        if y1_fill is not None and not isinstance(y1_fill, number_types):
            raise TypeError('Arg `y1_fill` is not a float or int.')
        if y2_fill is not None and not isinstance(y2_fill, number_types):
            raise TypeError('Arg `y2_fill` is not a float or int.')

        # Check arguments are valid.
        if x1 > x2:
            msg = f'Arg `x1=={x1}` is higher than `x2=={x2}`'
            raise ValueError(msg)

        # Copy some arguments to self for later use.
        self._x1 = x1
        self._x2 = x2
        self._y1_fill = y1_fill
        self._y2_fill = y2_fill

        # Calculate coefficients for the linear mapping.
        self._a = (y2 - y1) / (x2 - x1)
        self._b = y1 - self._a * x1

    def __call__(self, x):
        """
        :param x:
            Either a single float, int or other type of number,
            or a Numpy or Pandas array of numbers. Data-type is NOT checked!

        :return:
            Linear mapping of the same type as input `x`.
        """
        # Calculate the linear map for all input-values.
        y = self._a * x + self._b

        # Copy variables from self for easy reference.
        y1_fill = self._y1_fill
        y2_fill = self._y2_fill

        # Limit the values beyond the left boundary?
        if y1_fill is not None:
            # Boolean condition whether the input is beyond the boundary.
            # This works for both scalar values and arrays.
            cond = (x < self._x1)

            # Set the relevant output-numbers to the fill-number.
            if isinstance(x, Number):
                # Input is a single scalar number.
                if cond:
                    y = y1_fill
            else:
                # Input is assumed to be a Numpy or Pandas array of numbers.
                y[cond] = y1_fill

        # Limit the values beyond the right boundary?
        if self._y2_fill is not None:
            # Boolean condition whether the input is beyond the boundary.
            # This works for both scalar values and arrays.
            cond = (x > self._x2)

            # Set the relevant output-numbers to the fill-number.
            if isinstance(x, Number):
                if cond:
                    # Input is a single scalar number.
                    y = y2_fill
            else:
                # Input is assumed to be a Numpy or Pandas array of numbers.
                y[cond] = y2_fill

        return y


###############################################################################



###############################################################################
#
# Functions for diversifying an investment portfolio when the
# correlation matrix is sparse so it mostly contains zeros.
#
# This implements the sparse version of "Hvass Diversification" from the paper:
# - M.E.H. Pedersen, "Fast Portfolio Diversification", 2022.
#   https://ssrn.com/abstract=4009041
#   https://github.com/Hvass-Labs/Finance-Papers
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2022 by Magnus Erik Hvass Pedersen
#
###############################################################################

import numpy as np
import pandas as pd
from numba import jit
from investops.diversify import _check_convergence, _check_weights
from investops.sparse import sparse_corr_to_numpy

###############################################################################
# Helper functions.

def _fix_input(weights_org, corr_coef, weights_guess=None):
    """
    Ensure the weights and correlation coefficients do not have NaN-values
    (Not-a-Number) by filling with 0.

    This returns a copy of the input data.

    :param weights_org:
        Numpy array with the originally desired portfolio weights.

    :param corr_coef:
        Numpy array with the correlation coefficients of a sparse matrix.

    :param weights_guess:
        Numpy array with portfolio weights for the starting guess.

    :return:
        - New Numpy array with portfolio weights.
        - New Numpy array with the correlation coefficients of a sparse matrix.
        - New Numpy array with the initial portfolio weights guess.
    """
    # Copy arrays and fill NaN-values with 0.0
    weights_org = np.nan_to_num(weights_org, nan=0.0, copy=True)
    corr_coef = np.nan_to_num(corr_coef, nan=0.0, copy=True)
    if weights_guess is not None:
        weights_guess = np.nan_to_num(weights_guess, nan=0.0, copy=True)

    return weights_org, corr_coef, weights_guess


def _check_pandas_index(weights_org, weights_guess=None):
    """
    If the arguments are Pandas Series or DataFrames, then check that their
    indices have matching names, otherwise throw a `ValueError`.

    This is because Pandas can automatically align the data when doing math
    operations on the data, but we will be using Numpy in these algorithms,
    so the data would get corrupted if the index names don't match.

    :param weights_org:
        Array with the original asset-weights.

    :param weights_guess:
        Array with a starting guess for the diversified asset-weights.

    :raises:
        `ValueError` if the args have mis-matched Pandas index.

    :return:
        None
    """
    # Booleans whether the args are Pandas data-types.
    is_pandas_org = isinstance(weights_org, (pd.Series, pd.DataFrame))
    is_pandas_guess = isinstance(weights_guess, (pd.Series, pd.DataFrame))

    # Check weights_org and weights_guess.
    if is_pandas_org and is_pandas_guess:
        if not weights_org.index.equals(weights_guess.index):
            msg = 'Mismatch in index names for Pandas data.'
            raise ValueError(msg)

###############################################################################
# Full Exposure.

@jit(parallel=False)
def full_exposure_sparse(weights, corr_i, corr_j, corr_coef):
    """
    Calculate the so-called "Full Exposure" for each asset, which measures
    the entire portfolio's exposure to each asset both directly through the
    asset weights themselves, but also indirectly through their correlations
    with other assets in the portfolio.

    This function is decorated with Numba Jit, which means it compiles into
    super-fast machine-code the first time it is run. Note that this function
    currently does NOT support parallel execution.

    The sparse correlation matrix is given in Coordinate (COO) format
    with 3 arrays for the matrix coordinates and correlation coefficients.

    NOTE! The arguments must be either Numpy arrays or Python lists.
    The arguments cannot be Pandas Series or DataFrames, because Numba Jit
    does not support Pandas. The optimal run-speed is when using Numpy arrays.

    WARNING! It is very important that the sparse matrix ONLY contains the
    upper-triangle matrix, and it must NOT contain the diagonal and
    lower-triangle matrix, because the algorithm would then double-count
    the correlations and make faulty the calculations.

    The sparse algorithm for the Full Exposure is from the following paper:
    - M.E.H. Pedersen, "Fast Portfolio Diversification", 2022.
      https://ssrn.com/abstract=4009041
      https://github.com/Hvass-Labs/Finance-Papers

    :param weights:
        Numpy array with floats for the portfolio weights for the assets.

    :param corr_i:
        Numpy array with ints for the row indices of the sparse corr matrix.

    :param corr_j:
        Numpy array with ints for the column indices of the sparse corr matrix.

    :param corr_coef:
        Numpy array with floats for the correlation coefficients of the
        sparse correlation matrix.

    :returns:
        Numpy array with the Full Exposure of each asset.
    """
    # Initialize the array of correlated exposures with the squared weights,
    # because this is the minimum possible value for the sum of corr. exposures
    # if all the asset's correlation coefficients are zero.
    sum_corr_exp = np.copy(weights) ** 2

    # For each correlation coefficient in the sparse matrix.
    # WARNING! There is a "race condition" when this loop is run in parallel,
    # because the array sum_corr_exp is read and written by all the threads.
    for i, j, c in zip(corr_i, corr_j, corr_coef):
        # The indices i and j are for the row and column of the corr. matrix,
        # and c is the correlation coefficient at that position in the matrix.

        # Portfolio weights of assets i and j.
        w_i = weights[i]
        w_j = weights[j]

        # Product of the two asset weights and their correlation.
        prod = w_i * w_j * c

        # If the product is positive then the correlation is deemed "bad"
        # and must be included in the calculation of the Full Exposure,
        # so the two portfolio weights can be adjusted accordingly.
        # This is explained in Section 2.1 of the paper referenced above.
        if prod > 0.0:
            # Multiply with the correlation again, because otherwise the
            # square-root calculated below would amplify the correlation.
            # Because this can result in a negative number, we also need
            # to take the absolute value.
            prod_final = np.abs(prod * c)

            # Update the sum of correlation exposures for assets i and j.
            # Because the correlation is symmetrical, the two assets are
            # updated identically.
            sum_corr_exp[i] += prod_final
            sum_corr_exp[j] += prod_final

    # Finalize the calculation of the Full Exposure.
    full_exp = np.sign(weights) * np.sqrt(sum_corr_exp)

    return full_exp

###############################################################################
# Diversification algorithm.

@jit(parallel=False)
def _update_weights_sparse(weights_org, weights_new,
                           corr_i, corr_j, corr_coef):
    """
    Helper-function for the function `diversify_weights_sparse` which performs
    a single update of the portfolio weights, using the algorithm from
    Section 3.1 of the paper referenced above.

    The sparse correlation matrix is given in Coordinate (COO) format
    with 3 arrays for the matrix coordinates and correlation coefficients.

    NOTE! The arguments should be Numpy arrays for optimal speed.

    WARNING! It is very important that the sparse matrix ONLY contains the
    upper-triangle matrix, and it must NOT contain the diagonal and
    lower-triangle matrix, because the algorithm would then double-count
    the correlations and make faulty calculations.

    :param weights_org:
        Numpy array with floats for the original portfolio weights.

    :param weights_new:
        Numpy array with the adjusted portfolio weights. It is updated inplace.

    :param corr_i:
        Numpy array with ints for the row indices of the sparse corr matrix.

    :param corr_j:
        Numpy array with ints for the column indices of the sparse corr matrix.

    :param corr_coef:
        Numpy array with floats for the correlation coefficients of the
        sparse correlation matrix.

    :return:
        Float with the max absolute difference between the Full Exposure
        and the original portfolio weights. This is used to abort the outer
        algorithm's for-loop when sufficiently good weights have been found.
    """
    # This could be implemented entirely with Numpy, but would need special
    # handling in case the Full Exposure is zero for some assets. This is a
    # slightly cleaner implementation that is also faster with Numba jit.

    # Calculate the Full Exposure for all the portfolio weights.
    full_exp = full_exposure_sparse(weights=weights_new,
                                    corr_i=corr_i, corr_j=corr_j,
                                    corr_coef=corr_coef)

    # Init. max abs difference between the Full Exposure and original weights.
    max_abs_dif = 0.0

    # Number of portfolio weights.
    n = len(weights_org)

    # For each portfolio weight.
    # WARNING! There is a "race condition" when this loop is run in parallel,
    # because the variable max_abs_dif is read and written by all the threads.
    for i in range(n):
        # Original weight for asset i.
        w_org_i = weights_org[i]

        # Full Exposure for asset i.
        full_exp_i = full_exp[i]

        # If the Full Exposure is non-zero.
        if full_exp_i != 0.0:
            # Update the portfolio weight for asset i.
            weights_new[i] *= w_org_i / full_exp_i

        # Update max abs difference between Full Exposure and original weight.
        abs_dif = np.abs(full_exp_i - w_org_i)
        if abs_dif > max_abs_dif:
            max_abs_dif = abs_dif

    return max_abs_dif


def diversify_weights_sparse(weights_org, corr_i, corr_j, corr_coef,
                             weights_guess=None, fix_input=True,
                             log=None, max_iter=100, tol=1e-3):
    """
    Find new asset-weights that minimize the Mean Squared Error (MSE) between
    the original asset-weights and the Full Exposure of the new asset-weights.

    This is the sparse version of the algorithm, which is generally only faster
    than the dense version of the algorithm, if the correlation matrix is very
    sparse so most of the correlation coefficients are zero.
    See the paper referenced below for timing experiments.

    The sparse correlation matrix is given in Coordinate (COO) format
    with 3 arrays for the matrix coordinates and correlation coefficients.

    WARNING! It is very important that the sparse matrix ONLY contains the
    upper-triangle matrix, and it must NOT contain the diagonal and
    lower-triangle matrix, because the algorithm would then double-count
    the correlations and make faulty calculations.

    The sparse version of the diversification algorithm is from this paper:
    - M.E.H. Pedersen, "Fast Portfolio Diversification", 2022.
      https://ssrn.com/abstract=4009041
      https://github.com/Hvass-Labs/Finance-Papers

    :param weights_org:
        Array with the originally desired asset-weights for the portfolio.
        These can be either positive or negative, and they need not sum to 1.
        This data can either be a Pandas Series or Numpy array.

    :param corr_i:
        Array or list of the row indices of the sparse correlation matrix.
        If `weights_org` is a Numpy array, then this is a Numpy array of ints.
        If `weights_org` is a Pandas Series, then this is a list of strings.

    :param corr_j:
        Array or list of the column indices of the sparse correlation matrix.
        If `weights_org` is a Numpy array, then this is a Numpy array of ints.
        If `weights_org` is a Pandas Series, then this is a list of strings.

    :param corr_coef:
        Numpy array with floats for the correlation coefficients of the
        sparse correlation matrix.

    :param weights_guess:
        Array with a starting guess for the adjusted portfolio weights.
        If you are calling this function with `weights_org` and `corr` being
        nearly identical on each call, then you might save computation time
        by passing the last weights that were output from this function as
        the arg `weights_guess` the next time you call this function. This
        may reduce the number of iterations needed for convergence.

    :param fix_input:
        Boolean whether to repair input by filling NaN-values (Not-a-Number)
        in `weights_org`, `weights_guess` and `corr_coef`.

    :param log:
        If this is a list-like object then it will have its function `append`
        called after each iteration with the new weights, so you can print
        them later. This is useful for debugging and other demonstrations.

    :param max_iter:
        Max iterations of the algorithm.

    :param tol:
        Stop the algorithm when asset-weight adjustments are smaller than this
        tolerance level.

    :return:
        Array with the adjusted asset-weights.
        If `weights_org` is a Pandas Series or DataFrame,
        then this is also a Pandas Series, otherwise this is a Numpy array.
    """
    # If using Pandas data-types, ensure their index names match.
    # Note: This does NOT check the sparse correlation matrix,
    # which is tested during the conversion process below.
    _check_pandas_index(weights_org=weights_org, weights_guess=weights_guess)

    # Convert weights_org and sparse corr. matrix from Pandas to Numpy.
    if isinstance(weights_org, (pd.Series, pd.DataFrame)):
        # Save the Pandas index for later use with the return-data.
        index = weights_org.index

        # Convert the weights and sparse correlation matrix to Numpy arrays.
        # Args corr_i and corr_j are assumed to be lists of strings which must
        # be present in the index of weights_org, otherwise an error is raised.
        weights_org, corr_i, corr_j, corr_coef = \
            sparse_corr_to_numpy(weights=weights_org, corr_i=corr_i,
                                 corr_j=corr_j, corr_coef=corr_coef)
    else:
        # This is also used to indicate that the input was NOT Pandas data.
        index = None

        # Ensure the sparse correlation arrays are in Numpy format.
        # This may or may not be a copy of the data.
        corr_i = np.asarray(corr_i)
        corr_j = np.asarray(corr_j)
        corr_coef = np.asarray(corr_coef)

    # Convert weights_guess from Pandas to Numpy.
    if isinstance(weights_guess, (pd.Series, pd.DataFrame)):
        # This may or may not be a copy of the data.
        # Note: Flatten is necessary if it is a Pandas DataFrame.
        weights_guess = weights_guess.to_numpy().flatten()

    # Ensure the weights and correlation coefficients are valid.
    if fix_input:
        # This copies the data.
        weights_org, corr_coef, weights_guess = \
            _fix_input(weights_org=weights_org, corr_coef=corr_coef,
                       weights_guess=weights_guess)

    # Select a starting point for the new adjusted weights.
    # The arrays are copied so we don't modify the argument data.
    # It is possible that the data was already copied above, so there
    # is a slight redundancy here, but it makes the code easier to read.
    if weights_guess is not None:
        # Use the guessed weights as the starting point.
        # In case a guessed weight is zero, use the original weight,
        # otherwise the weight-adjustment would always get stuck in zero.
        # This should create a new numpy array so there is no need to copy.
        weights_new = np.where(weights_guess != 0.0, weights_guess, weights_org)
    else:
        # Use a copy of the original weights as the starting point.
        weights_new = np.copy(weights_org)

    # Log the initial weights?
    if log is not None:
        # Array is copied because the update iterates on the same array, so
        # the entire log would be filled with the same values if not copied.
        log.append(weights_new.copy())

    # Repeat for a number of iterations or until convergence
    # which breaks out of the for-loop further below.
    for i in range(max_iter):
        # Update the array weights_new inplace.
        max_abs_dif = _update_weights_sparse(weights_org=weights_org,
                                             weights_new=weights_new,
                                             corr_i=corr_i, corr_j=corr_j,
                                             corr_coef=corr_coef)

        # Log the updated weights?
        if log is not None:
            # Array is copied because the update iterates on the same array, so
            # the entire log would be filled with the same values if not copied.
            log.append(weights_new.copy())

        # Abort the for-loop when converged to a solution.
        if max_abs_dif < tol:
            break

    # Check that we have converged to a sufficiently good solution.
    _check_convergence(max_abs_dif=max_abs_dif, tol=tol)

    # Check that the original and new portfolio weights are consistent.
    _check_weights(weights_org=weights_org, weights_new=weights_new)

    # If the input weights_org was Pandas data, then also output Pandas data.
    if index is not None:
        weights_new = pd.Series(data=weights_new, index=index)

    return weights_new

###############################################################################
# Convert log.

def log_to_dataframe_sparse(weights_org, corr_i, corr_j, corr_coef, log):
    """
    Convert the log from `diversify_weights_sparse` to a Pandas DataFrame which
    shows the iterations of the adj. portfolio weights and their Full Exposure.

    The sparse correlation matrix is given in Coordinate (COO) format
    with 3 arrays for the matrix coordinates and correlation coefficients.

    WARNING! It is very important that the sparse matrix ONLY contains the
    upper-triangle matrix, and it must NOT contain the diagonal and
    lower-triangle matrix, because the algorithm would then double-count
    the correlations and make faulty calculations.

    :param weights_org:
        Numpy array with floats for the originally desired portfolio weights.

    :param corr_i:
        Numpy array with ints for the row indices of the sparse corr matrix.

    :param corr_j:
        Numpy array with ints for the column indices of the sparse corr matrix.

    :param corr_coef:
        Numpy array with floats for the correlation coefficients of the
        sparse correlation matrix.

    :param log:
        List of Numpy arrays with portfolio weights. This is obtained by first
        passing the list as the `log` arg in the `diversify_weights_sparse`
        function.

    :return:
        Pandas DataFrame
    """
    # Convert log to numpy 2-dim array.
    log_weights = np.array(log)

    # Get the number of iterations and assets in the log.
    num_iterations, num_assets = log_weights.shape

    # Initialize log for the Full Exposure.
    log_full_exp = []

    # Initialize log for the Mean Squared Error (MSE).
    log_mse = []

    # For each array of adjusted weights in the log.
    for weights_new in log_weights:
        # Calculate the Full Exposure of the logged weights.
        fe = full_exposure_sparse(weights=weights_new, corr_i=corr_i,
                                  corr_j=corr_j, corr_coef=corr_coef)

        # Save the results.
        log_full_exp.append(fe)

        # Calculate the Mean Squared Error (MSE).
        mse = np.mean((fe - weights_org) ** 2)

        # Save the results.
        log_mse.append(mse)

    # Combine the arrays of adjusted weights and Full Exposure, so that:
    # 1st column is for 1st weights, 2nd column is for 1st Full Exposure.
    # 3rd column is for 2nd weights, 4th column is for 2nd Full Exposure.
    data = np.dstack((log_weights, log_full_exp)).reshape(num_iterations, -1)

    # Generate names for the columns.
    names = []
    for i in range(1, num_assets + 1):
        names.append(f'Weight {i}')
        names.append(f'Full Exp. {i}')

    # Index for the rows.
    index = pd.Series(data=list(range(0, num_iterations)), name='Iteration')

    # Create Pandas DataFrame with the data.
    df = pd.DataFrame(data=data, columns=names, index=index)

    # Append a column for the Mean Squared Error (MSE).
    df['MSE'] = log_mse

    return df

###############################################################################



###############################################################################
#
# Functions for checking and repairing data.
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

import numpy as np
from numba import jit

###############################################################################


@jit
def fix_corr_matrix(corr, copy=True):
    """
    Repair a correlation matrix so it is symmetrical, limited between -1 and 1,
    and the diagonal elements are all 1. The upper-triangle is copied to the
    lower-triangle. NaN values are set to zero.

    :param corr:
        Numpy 2-dim array for the correlation matrix which is updated inplace.

    :param copy:
        Boolean whether to copy `corr` (True) or update `corr` inplace (False).

    :return:
        Numpy array with the repaired correlation matrix.
    """
    # Update the data inplace or make a copy?
    if copy:
        corr = corr.copy()

    # Number of rows and columns.
    n = len(corr)

    # For each row.
    for i in range(n):
        # For each column after the diagonal.
        for j in range(i + 1, n):
            # Get the correlation value.
            c = corr[i, j]

            #  Ensure the correlation value is valid.
            if np.isnan(c):
                # NaN (Not-a-Number) value is set to zero.
                c = 0.0
            elif c > 1.0:
                # Clip the value if it is higher than 1.0
                c = 1.0
            elif c < -1.0:
                # Clip the value if it is lower than -1.0
                c = -1.0

            # Update the matrix inplace. Also copy to the lower-triangle.
            corr[i, j] = corr[j, i] = c

        # Ensure the diagonal is 1.
        corr[i, i] = 1.0

    return corr


@jit
def check_corr_matrix(corr, tol=1e-9):
    """
    Check that a numpy array is a valid correlation matrix:

    - It must be matrix-shaped.
    - Its elements must be between -1 and 1.
    - The diagonal must be 1.
    - The matrix must be symmetrical.

    The checks allow for small floating point rounding errors.

    :param corr:
        Numpy 2-dim array for the correlation matrix.
        Note: It is NOT checked that it is a valid Numpy array, because that
        kind of type-checking is not supported inside a Numba Jit function.

    :param tol:
        Float with the error tolerance in the float comparisons.

    :raises:
        `ValueError` if the `corr` arg is an invalid correlation matrix.

    :return:
        None
    """
    # Assume `corr` is a valid Numpy array, because we cannot check its type
    # inside a Numba Jit function using e.g. isinstance(corr, np.ndarray).

    # Check it is matrix-shaped.
    if corr.ndim != 2 or corr.shape[0] != corr.shape[1]:
        raise ValueError('Correlation matrix is not matrix-shaped.')

    # Number of rows and columns.
    n = corr.shape[0]

    # For each row in the correlation matrix.
    for i in range(n):
        # Get diagonal of the matrix.
        corr_ii = corr[i, i]

        # Check the diagonal is not NaN (Not-a-Number) or Infinity.
        if np.isnan(corr_ii) or np.isinf(corr_ii):
            raise ValueError('Correlation matrix has NaN or Infinity.')

        # Check the diagonal is 1.
        if np.abs(corr_ii - 1.0) > tol:
            raise ValueError('Correlation matrix diagonal is not 1.')

        # For each column after the diagonal.
        for j in range(i + 1, n):
            # Get elements from the correlation matrix.
            corr_ij = corr[i, j]
            corr_ji = corr[j, i]

            # Check the correlations are not NaN (Not-a-Number) or Infinity.
            if np.isnan(corr_ij) or np.isinf(corr_ij):
                raise ValueError('Correlation matrix has NaN or Infinity.')

            # Check the correlations are between -1 and 1.
            if (corr_ij < -1.0 - tol) or (corr_ij > 1.0 + tol):
                msg = 'Correlation matrix has element outside range [-1,1].'
                raise ValueError(msg)

            # Check the matrix is symmetrical.
            if np.abs(corr_ij - corr_ji) > tol:
                raise ValueError('Correlation matrix is not symmetrical.')


###############################################################################



###############################################################################
#
# Functions for generating random portfolio weights and correlation matrices.
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

import numpy as np
from investops.check import fix_corr_matrix

###############################################################################
# Random Weights.

def rand_weights_uniform(rng, num_assets, min_weight=0.0, max_weight=1.0):
    """
    Generate random portfolio weights using a uniform distribution.

    :param rng: `Numpy.random.Generator` object from `np.random.default_rng()`
    :param num_assets: Integer with the number of portfolio assets.
    :param min_weight: Minimum portfolio weight.
    :param max_weight: Maximum portfolio weight.
    :return: Numpy array with random weights.
    """
    return rng.uniform(low=min_weight, high=max_weight, size=num_assets)


def rand_weights_normal(rng, num_assets, mean=0.0, std=0.04,
                        min_weight=0.0, max_weight=1.0):
    """
    Generate random portfolio weights using a normal distribution that
    are also clipped between the given `min_weight` and `max_weight`.

    :param rng: `Numpy.random.Generator` object from `np.random.default_rng()`
    :param num_assets: Integer with the number of portfolio assets.
    :param mean: Mean of the normal distribution.
    :param std: Std.dev. of the normal distribution.
    :param min_weight: Minimum portfolio weight.
    :param max_weight: Maximum portfolio weight.
    :return: Numpy array with random weights.
    """
    # Generate random portfolio weights.
    weights = rng.normal(scale=std, loc=mean, size=num_assets)

    # Clip the portfolio weights.
    weights = np.clip(weights, min_weight, max_weight)

    return weights


###############################################################################
# Random correlation matrices.

def rand_corr_uniform(rng, num_assets, min_corr=-1.0, max_corr=1.0):
    """
    Generate a random correlation matrix using a uniform distribution.
    It is symmetrical with random elements between `min_corr` and `max_corr`
    (also limited between -1 and 1), and the diagonal elements are all 1.

    :param rng: `Numpy.random.Generator` object from `np.random.default_rng()`
    :param num_assets: Integer for the number of portfolio assets.
    :param min_corr: Minimum correlation coefficient.
    :param max_corr: Maximum correlation coefficient.
    :return: Numpy matrix with random correlation coefficients.
    """
    # Generate random matrix.
    corr = rng.uniform(low=min_corr, high=max_corr,
                       size=(num_assets, num_assets))

    # Ensure it is a valid correlation matrix. Repair the data inplace.
    fix_corr_matrix(corr, copy=False)

    return corr


def rand_corr_normal(rng, num_assets, mean=0.0, std=0.2):
    """
    Generate a random correlation matrix using a normal distribution.
    It is symmetrical with random elements limited between -1 and 1,
    and the diagonal elements are all 1.

    :param rng: `Numpy.random.Generator` object from `np.random.default_rng()`
    :param num_assets: Integer for the number of portfolio assets.
    :param mean: Mean of the normal distribution.
    :param std: Std.dev. of the normal distribution.
    :return: Numpy matrix with random correlation coefficients.
    """
    # Generate random matrix.
    corr = rng.normal(scale=std, loc=mean, size=(num_assets, num_assets))

    # Ensure it is a valid correlation matrix. Repair the data inplace.
    fix_corr_matrix(corr, copy=False)

    return corr


###############################################################################


def rand_zero(rng, a, prob):
    """
    Randomly set values of a Numpy array to zero according to a probability.

    :param rng: `Numpy.random.Generator` object from `np.random.default_rng()`
    :param a: Numpy array which is copied and NOT updated inplace.
    :param prob: Float with the probability that an element is set to zero.
    :return: A new Numpy array of the same shame as `a`.
    """
    # Array with random values between [0,1].
    p = rng.uniform(size=a.shape)

    # Create a copy of the input array where random values are set to zero.
    return np.where(p < prob, 0, a)


###############################################################################



###############################################################################
#
# Functions for diversifying an investment portfolio.
#
# This implements so-called "Hvass Diversification" from the paper:
# - M.E.H. Pedersen, "Simple Portfolio Optimization That Works!", 2021.
#   https://ssrn.com/abstract=3942552
#   https://github.com/Hvass-Labs/Finance-Papers
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

import numpy as np
import pandas as pd
from numba import jit, prange
from investops.check import fix_corr_matrix

###############################################################################
# Helper functions.

def _fix_input(weights_org, corr, weights_guess=None):
    """
    Ensure the weights do not have NaN-values (Not-a-Number) by filling with 0.
    And make various repairs to the correlation-matrix so it is valid.

    This makes and returns a copy of the input data.

    :param weights_org:
        Numpy array with the originally desired portfolio weights.

    :param corr:
        Numpy 2-dim array with the correlation-matrix.

    :param weights_guess:
        Numpy array with portfolio weights for the starting guess.

    :return:
        New Numpy array with portfolio weights.
        New Numpy 2-dim array with the correlation matrix.
        New Numpy array with the initial portfolio weights guess.
    """
    # Copy arrays and fill NaN-values with 0.0
    weights_org = np.nan_to_num(weights_org, nan=0.0, copy=True)
    if weights_guess is not None:
        weights_guess = np.nan_to_num(weights_guess, nan=0.0, copy=True)

    # Repair the correlation-matrix. This makes a copy of the data.
    corr = fix_corr_matrix(corr=corr, copy=True)

    return weights_org, corr, weights_guess


def _check_pandas_index(weights_org, corr, weights_guess=None):
    """
    If the arguments are Pandas Series or DataFrames, then check that their
    indices and columns have matching names, otherwise throw a `ValueError`.

    This is because Pandas can automatically align the data when doing math
    operations on the data, but we will be using Numpy in these algorithms,
    so the data would get corrupted if index and column names don't match.

    The time-usage is around 0.1 milli-seconds when `len(weights_org) == 1000`.

    :param weights_org:
        Array with the original asset-weights.

    :param corr:
        Matrix with the correlations between assets.

    :param weights_guess:
        Array with a starting guess for the diversified asset-weights.

    :raises:
        `ValueError` if the args have mis-matched Pandas index and column.

    :return:
        None
    """
    # Booleans whether the args are Pandas data-types.
    is_pandas_org = isinstance(weights_org, (pd.Series, pd.DataFrame))
    is_pandas_corr = isinstance(corr, pd.DataFrame)
    is_pandas_guess = isinstance(weights_guess, (pd.Series, pd.DataFrame))

    # Error message.
    msg = 'Mismatch in index / column names for Pandas data.'

    # Check weights_org and corr.
    if is_pandas_org and is_pandas_corr:
        if not (weights_org.index.equals(corr.index) and
                weights_org.index.equals(corr.columns)):
            raise ValueError(msg)

    # Check weights_org and weights_guess.
    if is_pandas_org and is_pandas_guess:
        if not weights_org.index.equals(weights_guess.index):
            raise ValueError(msg)

    # Check weights_guess and corr.
    # This is only necessary if weights_org is not a Pandas data-type,
    # otherwise we would already know that weights_org matches corr and
    # weights_org matches weights_guess, therefore weights_guess matches corr.
    if (not is_pandas_org) and is_pandas_guess and is_pandas_corr:
        if not (weights_guess.index.equals(corr.index) and
                weights_guess.index.equals(corr.columns)):
            raise ValueError(msg)


@jit
def _find_weight_problem(weights_org, weights_new):
    """
    Helper-function for the `_check_weights` function which returns the index
    of the first problem for the portfolio weights. Runs fast with Numba Jit.

    :param weights_new:
        Numpy array with the new asset-weights.

    :param weights_org:
        Numpy array with the original asset-weights.

    :return:
        `None` if no problems were found.
         Otherwise an integer with the index of the first problem.
    """
    # Number of weights.
    n = len(weights_new)

    # For each weight index.
    for i in range(n):
        # Get the weights.
        w_new = weights_new[i]
        w_org = weights_org[i]

        # Check if there is a problem and then return the corresponding index.
        # We must ensure the weight signs are equal and magnitudes are valid.
        # But because np.sign(0.0)==0.0 the check for signs is a bit awkward.
        if (np.sign(w_new) != 0.0 and np.sign(w_new) != np.sign(w_org)) or \
                (np.abs(w_new) > np.abs(w_org)):
            return i

    # No problems were found.
    return None


def _check_weights(weights_org, weights_new):
    """
    Check that the original and new portfolio weights are consistent. They must
    have the same sign, and the absolute values of the new weights must be
    smaller than the absolute values of the original weights:

    (1)     sign(weights_new[i]) == sign(weights_org[i])
    (2)     abs(weights_new[i]) <= abs(weights_org[i])

    This function only takes 3.5 micro-seconds to run for 1000 weights using a
    Numba Jit implementation. A Numpy implementation would be much slower. But
    it must be split into two functions, because Numba Jit does not properly
    support the string operations used to generate the exception.

    :param weights_new:
        Numpy array with the new asset-weights.

    :param weights_org:
        Numpy array with the original asset-weights.

    :raises:
        `RuntimeError` if the weights are inconsistent.

    :return:
        None
    """
    # Get index of the first problem / inconsistency of the weights.
    idx = _find_weight_problem(weights_org=weights_org,
                               weights_new=weights_new)

    # If a problem was found then raise an exception.
    if idx is not None:
        msg = f'Checking the weights failed at: i={idx}, ' + \
              f'weights_new[i]={weights_new[idx]:.2e}, ' + \
              f'weights_org[i]={weights_org[idx]:.2e}'
        raise RuntimeError(msg)


def _check_convergence(max_abs_dif, tol):
    """
    Check the adjusted portfolio weights have converged, so the Full Exposure
    of the portfolio weights are sufficiently close to the original weights.

    :param max_abs_dif:
        Float with max absolute difference between the Full Exposure and the
        original portfolio weights.

    :param tol:
        Tolerance level for the max abs difference.

    :raises:
        `RuntimeError` if the weights did not converge to the Full Exposure.

    :return:
        None
    """
    if max_abs_dif > tol:
        msg = 'Weights did not converge: ' + \
              f'max_abs_dif={max_abs_dif:.2e}, tol={tol:.2e}'
        raise RuntimeError(msg)


###############################################################################
# Full Exposure.

@jit(parallel=False)
def full_exposure(weights, corr):
    """
    Calculate the so-called "Full Exposure" for each asset, which measures
    the entire portfolio's exposure to each asset both directly through the
    asset weights themselves, but also indirectly through their correlations
    with other assets in the portfolio.

    There are different ways of defining the Full Exposure, as explained in
    the paper referenced below. This particular formula is Eq.(38) in that
    paper, which was found to work well in practice.

    The function is decorated with Numba Jit, which means it compiles into
    super-fast machine-code the first time it is run. This function is the most
    expensive part of the diversification method because it has time-complexity
    O(n^2) where n is the number of assets in the portfolio. Implementing it
    with for-loops instead of Numpy arrays, means that it avoids new memory
    allocations for large n^2 matrices, so the machine-code is very fast.

    For large portfolios of e.g. 1000 assets or more, this can run even faster
    by using `@jit(parallel=True)` as the function decorator. But for smaller
    portfolios of only e.g. 100 assets, the parallelization overhead makes it
    run a bit slower, so you have to turn on the parallelism manually.

    Note that the arguments must be Python lists or Numpy arrays and cannot be
    Pandas Series and DataFrames, because Numba Jit does not support Pandas.

    The Full Exposure is explained in more detail in the following paper:
    - M.E.H. Pedersen, "Simple Portfolio Optimization That Works!", 2021.
      https://ssrn.com/abstract=3942552
      https://github.com/Hvass-Labs/Finance-Papers

    :param weights:
        Numpy array with the portfolio weights for the assets.

    :param corr:
        Numpy 2-dim array with the correlation matrix for the assets.
        The element in the i'th row and j'th column is the correlation
        between assets i and j.

    :returns:
        Numpy array with the Full Exposure of each asset.
    """
    # Number of assets in the portfolio.
    n = len(weights)

    # Initialize an empty array for the results.
    full_exp = np.empty(shape=n, dtype=np.float64)

    # For each asset i in the portfolio.
    # Note the use of prange() instead of range() which instructs Numba Jit
    # to parallelize this loop, but only if @jit(parallel=True) was used,
    # otherwise this just becomes the ordinary Python range().
    for i in prange(n):
        # Portfolio weight of asset i.
        w_i = weights[i]

        # Initialize the sum of correlated exposures.
        sum_corr_exp = 0.0

        # For each other asset j in the portfolio.
        for j in range(n):
            # Portfolio weight of asset j.
            w_j = weights[j]

            # Correlation between assets i and j.
            c = corr[i, j]

            # Product of the two asset weights and their correlation.
            prod = w_i * w_j * c

            # If the product is positive then the correlation is deemed "bad"
            # and must be included in the calculation of the Full Exposure,
            # so the two portfolio weights can be adjusted accordingly.
            # This is explained in Section 8.3 of the paper referenced above.
            if prod > 0.0:
                # Multiply with the correlation again, because otherwise the
                # square-root calculated below would amplify the correlation.
                # Because this can result in a negative number, we also need
                # to take the absolute value.
                sum_corr_exp += np.abs(prod * c)

        # Calculate and save the Full Exposure for asset i.
        full_exp[i] = np.sign(w_i) * np.sqrt(sum_corr_exp)

    return full_exp


###############################################################################
# Mean Squared Error.

def mse_full_exposure(weights_new, weights_org, corr):
    """
    Mean Squared Error (MSE) between the original asset-weights
    and the Full Exposure of the new asset-weights.

    When the MSE value is zero, it means that the Full Exposure of
    the new asset-weights are equal to the original asset-weights.

    :param weights_org:
        Numpy array with the original asset-weights.

    :param weights_new:
        Numpy array with the new asset-weights.

    :param corr:
        Numpy 2-dim array for the correlation matrix between assets.

    :return:
        Float with the MSE value.
    """
    # Calculate the Full Exposure of the new asset-weights.
    full_exp = full_exposure(weights=weights_new, corr=corr)

    # Calculate the Mean Squared Error.
    mse = np.mean((full_exp - weights_org) ** 2)

    return mse


###############################################################################
# Adjust portfolio weights using custom algorithm.

@jit(parallel=False)
def _update_weights(weights_org, weights_new, corr):
    """
    Helper-function for the function `diversify_weights` which performs
    a single update of the portfolio weights, using the algorithm from
    Section 8.8 of the paper referenced above.

    WARNING: This should NOT be run in parallel with Numba Jit because there
    is a "race condition" in the for-loop that would corrupt the results.

    :param weights_org:
        Numpy array with the original portfolio weights.

    :param weights_new:
        Numpy array with the adjusted portfolio weights. It is updated inplace.

    :param corr:
        Numpy 2-dim array with the correlation matrix.

    :return:
        Float with the max absolute difference between the Full Exposure
        and the original portfolio weights. This is used to abort the outer
        algorithm's for-loop when sufficiently good weights have been found.
    """
    # This could be implemented entirely with Numpy, but would need special
    # handling in case the Full Exposure is zero for some assets. This is a
    # slightly cleaner implementation that is also faster with Numba jit.

    # Calculate the Full Exposure for all the portfolio weights.
    # This function can be run in parallel if enabled by the function.
    full_exp = full_exposure(weights=weights_new, corr=corr)

    # Init. max abs difference between the Full Exposure and original weights.
    max_abs_dif = 0.0

    # Number of portfolio weights.
    n = len(weights_org)

    # For each portfolio weight.
    # WARNING! There is a "race condition" when this loop is run in parallel,
    # because the variable max_abs_dif is read and written by all the threads.
    for i in range(n):
        # Original weight for asset i.
        w_org_i = weights_org[i]

        # Full Exposure for asset i.
        full_exp_i = full_exp[i]

        # If the Full Exposure is non-zero.
        if full_exp_i != 0.0:
            # Update the portfolio weight for asset i.
            weights_new[i] *= w_org_i / full_exp_i

        # Update max abs difference between Full Exposure and original weight.
        abs_dif = np.abs(full_exp_i - w_org_i)
        if abs_dif > max_abs_dif:
            max_abs_dif = abs_dif

    return max_abs_dif


def diversify_weights(weights_org, corr, weights_guess=None, fix_input=True,
                      log=None, max_iter=100, tol=1e-3):
    """
    Find new asset-weights that minimize the Mean Squared Error (MSE) between
    the original asset-weights and the Full Exposure of the new asset-weights,
    using a custom and extremely fast algorithm for this particular problem.

    For a portfolio of 1000 assets it only takes 20 milli-seconds to run this,
    depending on the CPU, arguments, and the weights and correlation matrix.
    Removing some of the options and overhead in the function can significantly
    improve the speed. But Numba Jit cannot improve the speed of this function.

    The diversification algorithm is explained in Section 8 of the paper:
    - M.E.H. Pedersen, "Simple Portfolio Optimization That Works!", 2021.
      https://ssrn.com/abstract=3942552
      https://github.com/Hvass-Labs/Finance-Papers

    :param weights_org:
        Array with the originally desired asset-weights for the portfolio.
        These can be either positive or negative and they need not sum to 1.
        This data can either be a Pandas Series or Numpy array.

    :param corr:
        Matrix with the correlations between assets.
        This can either be a Pandas DataFrame or Numpy array.

    :param weights_guess:
        Array with a starting guess for the adjusted portfolio weights.
        If you are calling this function with `weights_org` and `corr` being
        nearly identical on each call, then you might save computation time
        by passing the last weights that were output from this function as
        the arg `weights_guess` the next time you call this function. This
        may reduce the number of iterations needed for convergence.

    :param fix_input:
        Boolean whether to repair input by filling NaN-values (Not-a-Number)
        in `weights_org` and `weights_guess`, and various repairs to `corr`.

    :param log:
        If this is a list-like object then it will have its function `append`
        called after each iteration with the new weights, so you can print
        them later. This is useful for debugging and other demonstrations.

    :param max_iter:
        Max iterations of the algorithm.

    :param tol:
        Stop the algorithm when asset-weight adjustments are smaller than this
        tolerance level.

    :return:
        Array with the adjusted asset-weights.
        If `weights_org` is a Pandas Series or DataFrame,
        then this is also a Pandas Series, otherwise this is a Numpy array.
    """
    # If using Pandas data-types, ensure their index and column names match.
    _check_pandas_index(weights_org=weights_org, corr=corr,
                        weights_guess=weights_guess)

    # Convert weights_org from Pandas to Numpy.
    if isinstance(weights_org, (pd.Series, pd.DataFrame)):
        # Save the Pandas index for later use with the return-data.
        index = weights_org.index

        # Convert Pandas to Numpy. This may or may not be a copy of the data.
        # Note: Flatten is necessary if it is a Pandas DataFrame.
        weights_org = weights_org.to_numpy().flatten()
    else:
        # This is also used to indicate that the input was NOT Pandas data.
        index = None

    # Convert weights_guess from Pandas to Numpy.
    if isinstance(weights_guess, (pd.Series, pd.DataFrame)):
        # This may or may not be a copy of the data.
        # Note: Flatten is necessary if it is a Pandas DataFrame.
        weights_guess = weights_guess.to_numpy().flatten()

    # Convert correlation matrix from Pandas to Numpy.
    if isinstance(corr, pd.DataFrame):
        # This may or may not be a copy of the data.
        corr = corr.to_numpy()

    # Ensure the weights and correlation-matrix are valid.
    if fix_input:
        # This copies the data.
        weights_org, corr, weights_guess = \
            _fix_input(weights_org=weights_org, corr=corr,
                       weights_guess=weights_guess)

    # Select a starting point for the new adjusted weights.
    # The arrays are copied so we don't modify the argument data.
    # It is possible that the data was already copied above, so there
    # is a slight redundancy here, but it makes the code easier to read.
    if weights_guess is not None:
        # Use the guessed weights as the starting point.
        # In case a guessed weight is zero, use the original weight,
        # otherwise the weight-adjustment would always get stuck in zero.
        # This should create a new numpy array so there is no need to copy.
        weights_new = np.where(weights_guess != 0.0, weights_guess, weights_org)
    else:
        # Use a copy of the original weights as the starting point.
        weights_new = np.copy(weights_org)

    # Log the initial weights?
    if log is not None:
        # Array is copied because the update iterates on the same array, so
        # the entire log would be filled with the same values if not copied.
        log.append(weights_new.copy())

    # Repeat for a number of iterations or until convergence
    # which breaks out of the for-loop further below.
    for i in range(max_iter):
        # Update the array weights_new inplace.
        max_abs_dif = _update_weights(weights_org=weights_org,
                                      weights_new=weights_new, corr=corr)

        # Log the updated weights?
        if log is not None:
            # Array is copied because the update iterates on the same array, so
            # the entire log would be filled with the same values if not copied.
            log.append(weights_new.copy())

        # Abort the for-loop when converged to a solution.
        if max_abs_dif < tol:
            break

    # Check that we have converged to a sufficiently good solution.
    _check_convergence(max_abs_dif=max_abs_dif, tol=tol)

    # Check that the original and new portfolio weights are consistent.
    _check_weights(weights_org=weights_org, weights_new=weights_new)

    # If the input weights_org was Pandas data, then also output Pandas data.
    if index is not None:
        weights_new = pd.Series(data=weights_new, index=index)

    return weights_new


###############################################################################
# Convert log.

def log_to_dataframe(weights_org, corr, log):
    """
    Convert the log from `diversify_weights` to a Pandas DataFrame which shows
    the iterations of the adjusted portfolio weights and their Full Exposure.

    :param weights_org:
        Numpy array with the originally desired portfolio weights.

    :param corr:
        Numpy 2-dim array with the correlation matrix.

    :param log:
        List of numpy arrays with portfolio weights. This is obtained by first
        passing the list as the `log` arg in the `diversify_weights` function.

    :return:
        Pandas DataFrame
    """
    # Convert log to numpy 2-dim array.
    log_weights = np.array(log)

    # Get the number of iterations and assets in the log.
    num_iterations, num_assets = log_weights.shape

    # Initialize log for the Full Exposure.
    log_full_exp = []

    # Initialize log for the Mean Squared Error (MSE).
    log_mse = []

    # For each array of adjusted weights in the log.
    for weights_new in log_weights:
        # Calculate the Full Exposure of the logged weights.
        fe = full_exposure(weights=weights_new, corr=corr)

        # Save the results.
        log_full_exp.append(fe)

        # Calculate the Mean Squared Error (MSE).
        mse = np.mean((fe - weights_org) ** 2)

        # Save the results.
        log_mse.append(mse)

    # Combine the arrays of adjusted weights and Full Exposure, so that:
    # 1st column is for 1st weights, 2nd column is for 1st Full Exposure.
    # 3rd column is for 2nd weights, 4th column is for 2nd Full Exposure.
    data = np.dstack((log_weights, log_full_exp)).reshape(num_iterations, -1)

    # Generate names for the columns.
    names = []
    for i in range(1, num_assets + 1):
        names.append(f'Weight {i}')
        names.append(f'Full Exp. {i}')

    # Index for the rows.
    index = pd.Series(data=list(range(0, num_iterations)), name='Iteration')

    # Create Pandas DataFrame with the data.
    df = pd.DataFrame(data=data, columns=names, index=index)

    # Append a column for the Mean Squared Error (MSE).
    df['MSE'] = log_mse

    return df


###############################################################################



###############################################################################
#
# Functions for removing zero/small portfolio weights and their correlations.
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

import pandas as pd
import numpy as np

###############################################################################
# Common helper-functions.


def _check_type(weights, corr=None):
    """
    Check that the arguments have the correct type.

    :param weights: Pandas Series with the portfolio weights.
    :param corr: Pandas DataFrame with the correlation matrix.
    :raises: `TypeError` if the args have incorrect types.
    :return: None
    """
    # Check the type of the weights.
    if not isinstance(weights, pd.Series):
        msg = 'Argument \'weights\' must be a Pandas Series.'
        raise TypeError(msg)

    # Check the type of the correlation matrix.
    if corr is not None and not isinstance(corr, pd.DataFrame):
        msg = 'Argument \'corr\' must be a Pandas DataFrame.'
        raise TypeError(msg)


def _remove_weights(mask, weights, corr=None):
    """
    Helper-function for removing portfolio weights according to a boolean mask.

    :param mask: Boolean mask for the weights to keep (True) or remove (False).
    :param weights: Pandas Series with the portfolio weights.
    :param corr: Pandas DataFrame with the correlation matrix.
    :return:
        If arg `corr` is not `None` then return a tuple with the new
        weights and correlations. Otherwise, just return the new weights.
    """

    # Only use the weights where the boolean mask is True.
    weights_new = weights[mask]

    if corr is not None:
        # Use the index of the weights (e.g. stock-tickers) instead of the
        # boolean mask, because then it also works for correlation matrices
        # that have more assets than just those in the weights. This also
        # sorts the rows and columns of the corr matrix to match the weights.
        index = weights_new.index

        # Only use the relevant elements of the correlation matrix.
        corr_new = corr.loc[index, index]

        # Return both the new weights and correlations.
        return weights_new, corr_new
    else:
        # Only return the new weights.
        return weights_new


###############################################################################


def remove_zero_weights(weights, corr=None):
    """
    Remove the portfolio weights that are zero (both 0.0 and -0.0),
    and also remove their correlations if the arg `corr` is not `None`.

    If you also want to remove weights that are `NaN` (Not-a-Number) then you
    call the function `dropna()` on the `weights` before calling this function.

    :param weights: Pandas Series with the portfolio weights.
    :param corr: Pandas DataFrame with the correlation matrix.
    :return:
        If arg `corr` is not `None` then return a tuple with the new
        weights and correlations. Otherwise, just return the new weights.
    """
    # Check the arg types are correct.
    _check_type(weights=weights, corr=corr)

    # Boolean mask for keeping only the non-zero portfolio weights.
    mask = (weights != 0.0)

    return _remove_weights(mask=mask, weights=weights, corr=corr)


def remove_low_weights(weights, corr=None, threshold=0.0):
    """
    Remove the portfolio weights whose absolute value is below the threshold,
    and also remove their correlations if the arg `corr` is not `None`.

    If you also want to remove weights that are `NaN` (Not-a-Number) then you
    call the function `dropna()` on the `weights` before calling this function.

    :param weights: Pandas Series with the portfolio weights.
    :param corr: Pandas DataFrame with the correlation matrix.
    :param threshold: Positive float with threshold for the portfolio weights.
    :return:
        If arg `corr` is not `None` then return a tuple with the new
        weights and correlations. Otherwise, just return the new weights.
    """
    # Check the arg types are correct.
    _check_type(weights=weights, corr=corr)

    # Check other args.
    if threshold < 0.0:
        msg = 'Argument \'threshold\' must be a positive number.'
        raise ValueError(msg)

    # Boolean mask for keeping only the relevant portfolio weights.
    mask = (np.abs(weights) >= threshold)

    return _remove_weights(mask=mask, weights=weights, corr=corr)


###############################################################################



###############################################################################
#
# Functions for normalizing portfolio weights.
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

import numpy as np
import pandas as pd
import numba
from numba import jit

###############################################################################
# Common helper-functions.

def _convert_input_weights(weights):
    """
    Convert the input weights from Numpy or Pandas to a 2-dim Numpy array.

    :param weights:
        Array with portfolio weights. Either a Numpy Array, Pandas Series, or
        Pandas DataFrame. Rows are for time-steps, columns are for assets.

    :raises:
        `TypeError` if the `weights` arg is of the wrong data-type.

    :return:
        Numpy 2-dim array with portfolio weights.
    """
    # Convert the weights data to a 2-dim Numpy array.
    if isinstance(weights, pd.Series):
        # Convert Pandas Series to 2-dim Numpy array.
        weights_np = weights.to_numpy()[np.newaxis, :]
    elif isinstance(weights, pd.DataFrame):
        # Convert Pandas DataFrame to 2-dim Numpy array.
        weights_np = weights.to_numpy()
    elif isinstance(weights, np.ndarray):
        if len(weights.shape) == 1:
            # Convert Numpy 1-dim array to 2-dim Numpy array.
            weights_np = weights[np.newaxis, :]
        else:
            # Use the Numpy 2-dim array as it is.
            weights_np = weights
    else:
        # Raise exception for invalid data-type.
        msg = 'Argument `weights` must be either Numpy Array, ' \
              'Pandas Series or DataFrame.'
        raise TypeError(msg)

    return weights_np


def _convert_output_weights(weights, weights_norm, cash):
    """
    Convert the normalized weights to the same data-type as the input weights.

    :param weights:
        Array with the input weights, only used to infer its data-type.

    :param weights_norm:
        Numpy 2-dim array with the normalized weights.

    :param cash:
        Numpy 1-dim array with the cash-positions.

    :return:
        Tuple with the normalized weights and cash as the proper data-types.
    """
    # Convert the result to the same data-type as the input.
    if isinstance(weights, pd.Series):
        # Convert Numpy 2-dim array to Pandas Series.
        weights_norm = pd.Series(data=weights_norm[0], index=weights.index)
        # Convert Numpy 1-dim array to scalar float.
        cash = cash[0]
    elif isinstance(weights, pd.DataFrame):
        # Convert Numpy 2-dim array to Pandas DataFrame.
        weights_norm = pd.DataFrame(data=weights_norm, index=weights.index,
                                    columns=weights.columns)
        # Convert Numpy 1-dim array to Pandas Series.
        cash = pd.Series(data=cash, index=weights.index)
    elif isinstance(weights, np.ndarray) and len(weights.shape) == 1:
        # Convert Numpy 2-dim array to Numpy 1-dim array.
        weights_norm = weights_norm[0, :]
        # Convert Numpy 1-dim array to scalar float.
        cash = cash[0]

    return weights_norm, cash


###############################################################################
# Functions for normalizing only positive weights.

@jit
def _normalize_weights_pos_jit(weights, limit_pos):
    """
    Fast implementation of the function `normalize_weights_pos` with Numba Jit.

    It is possible to implement this directly using Pandas, but even with a
    clever implementation it is harder to understand and runs much slower.

    :param weights:
        Numpy array with portfolio weights. This MUST be 2-dimensional!
        Rows are for time-steps, columns are for assets.

    :param limit_pos:
        Float with the maximum allowed sum of positive weights.

    :return:
        Tuple with the following data:
        - Numpy array 2-dim with the normalized weights.
        - Numpy array 1-dim with the cash-position for each time-step.
    """
    # Number of rows (time-steps) and number of columns (assets).
    num_rows, num_cols = weights.shape

    # Initialize the arrays for the normalized weights and cash-position.
    weights_norm = weights.copy()
    cash = np.zeros(num_rows, dtype=numba.float32)

    # For each row (time-step).
    for i in range(num_rows):
        # Initialize the sum of positive and negative weights for this row.
        weight_sum = 0.0

        # For each column (asset).
        for j in range(num_cols):
            # Get the weight for this row and column.
            w_ij = weights[i, j]

            # Accumulate the sum of weights for this row.
            # Note: We do NOT check if the weights are positive!
            weight_sum += w_ij

        if weight_sum > limit_pos:
            # Scale for all weights in this row to bring their sum down.
            scale = limit_pos / weight_sum

            # For each column (asset).
            for j in range(num_cols):
                # Update the weight with the scale.
                weights_norm[i, j] *= scale
        else:
            # This row was not scaled.
            scale = 1.0

        # Calculate and set the cash-position for this row (time-step).
        # Note: This can be negative if limit_pos > 1.0
        cash[i] = 1.0 - scale * weight_sum

    return weights_norm, cash


def normalize_weights_pos(weights, limit_pos=1.0):
    """
    Normalize portfolio weights so their sum does not exceed some limit. This
    function is only intended for positive (aka. "long") portfolio weights,
    and for maximum speed it does NOT check if the input weights are positive!

    If you are not sure whether some weights might be negative, you either need
    to clip the input-data to ensure it is positive, or you should use the
    other function `normalize_weights` which can also handle negative weights.

    This function is actually only about 20% faster than `normalize_weights`.

    :param weights:
        Array with portfolio weights. Either a Numpy Array, Pandas Series, or
        Pandas DataFrame. Rows are for time-steps, columns are for assets.

    :param limit_pos:
        Float with the maximum allowed sum of positive weights. Examples:
        Set this to 1.0 if you do not want to invest for borrowed money.
        Or set this to 0.8 if you only want to invest 80% of your money.
        Or set this to 1.2 if you want to invest for 20% borrowed money.

    :return:
        Tuple with the following data:
        - Array with the normalized weights. Same type and shape as `weights`.
        - Cash-position for each time-step, which can be negative if
          `limit_pos > 1.0` because you are investing for borrowed money.
          If `weights` is Pandas DataFrame, then this is a Pandas Series.
          If 'weights' is Numpy 2-dim array, then this is a 1-dim array.
          If `weights` is a Pandas Series or Numpy 1-dim array,
          then this is a float value.
    """
    # Convert input weights to Numpy 2-dim array.
    weights_np = _convert_input_weights(weights)

    # Run the Jit-optimized function to normalize the weights.
    weights_norm, cash = _normalize_weights_pos_jit(weights=weights_np,
                                                    limit_pos=limit_pos)

    # Convert output weights to same data-type as input weights.
    weights_norm, cash = _convert_output_weights(weights=weights, cash=cash,
                                                 weights_norm=weights_norm)

    return weights_norm, cash


###############################################################################
# Functions for normalizing both positive and negative weights.

@jit
def _normalize_weights_jit(weights, limit_pos, limit_neg, max_ratio):
    """
    Fast implementation of the function `normalize_weights` with Numba Jit.

    It is possible to implement this directly using Pandas, but even with a
    clever implementation, it is much harder to understand the code, and it
    runs about 1000x slower! The only drawback to this implementation, is
    that Numba Jit only supports Numpy arrays, so we must split the function
    into two parts: This function that does the actual computation, and another
    wrapper-function that can also handle Pandas Series and DataFrames.

    :param weights:
        Numpy array with portfolio weights. This MUST be 2-dimensional!
        Rows are for time-steps, columns are for assets.

    :param limit_pos:
        Float with the maximum allowed sum of positive weights.

    :param limit_neg:
        Float with the minimum allowed sum of negative weights.

    :param max_ratio:
        Float with maximum allowed ratio between positive and negative weights.

    :return:
        Tuple with the following data:
        - Numpy array 2-dim with the normalized weights.
        - Numpy array 1-dim with the cash-position for each time-step.
    """
    # Number of rows (time-steps) and number of columns (assets).
    num_rows, num_cols = weights.shape

    # Initialize the arrays for the normalized weights and cash-position.
    weights_norm = weights.copy()
    cash = np.zeros(num_rows, dtype=numba.float32)

    # For each row (time-step).
    for i in range(num_rows):
        # Initialize the sums of positive and negative weights for this row.
        sum_pos = 0.0
        sum_neg = 0.0

        # For each column (asset).
        for j in range(num_cols):
            # Get the weight for this row and column.
            w_ij = weights[i, j]

            # Accumulate sums of positive and negative weights for this row.
            if w_ij >= 0.0:
                sum_pos += w_ij
            else:
                sum_neg += w_ij

        # Scale used if the sum of positive weights exceeds its limit.
        if sum_pos > limit_pos:
            scale_pos = limit_pos / sum_pos
        else:
            scale_pos = 1.0

        # Scale used if the sum of negative weights exceeds its limit.
        if sum_neg < limit_neg:
            scale_neg = limit_neg / sum_neg
        else:
            scale_neg = 1.0

        # Adjust the scale for negative weights, if the sum of negative weights
        # exceed the sum of positive weights by the given ratio.
        if -sum_neg * scale_neg > max_ratio * sum_pos * scale_pos:
            scale_neg = max_ratio * sum_pos * scale_pos / -sum_neg

        # Scale this row if necessary.
        if scale_pos != 1.0 or scale_neg != 1.0:
            # For each column (asset).
            for j in range(num_cols):
                # Use the positive or negative scale for this weight.
                scale = scale_pos if weights[i, j] >= 0.0 else scale_neg

                # Update the weight with the scale.
                weights_norm[i, j] *= scale

        # Calculate and set the cash-position for this row (time-step).
        # Note: This can be negative if limit_pos > 1.0
        cash[i] = 1.0 - scale_pos * sum_pos

    return weights_norm, cash


def normalize_weights(weights, limit_pos=1.0, limit_neg=0.0, max_ratio=0.0):
    """
    Normalize portfolio weights so their sum does not exceed some limit.
    This can be used for "long", "short" and both "long and short" portfolios.

    For each time-step, the sums are calculated independently for the positive
    and negative weights. If the sums exceed their limits, then the weights are
    scaled to bring their sums back to their limits. Furthermore, the sum of
    negative weights is not allowed to exceed the sum of positive weights by
    the given `max_ratio`.

    NOTE: You should check that this normalization is consistent with your
    investment broker's requirements for "short" vs. "long" positions.

    :param weights:
        Array with portfolio weights. Either a Numpy Array, Pandas Series, or
        Pandas DataFrame. Rows are for time-steps, columns are for assets.

    :param limit_pos:
        Float with the maximum allowed sum of positive weights. Examples:
        Set this to 1.0 if you do not want to invest for borrowed money.
        Or set this to 0.8 if you only want to invest 80% of your money.
        Or set this to 1.2 if you want to invest for 20% borrowed money.

    :param limit_neg:
        Float with the minimum allowed sum of negative weights. Example:
        Set this to -0.5 if you only want to "short" for 50% of your money.

    :param max_ratio:
        Float with maximum allowed ratio between positive and negative weights.
        Example: Set this to 0.2 if the sum of negative weights must not exceed
        20% of the sum of positive weights.

    :return:
        Tuple with the following data:
        - Array with the normalized weights. Same type and shape as `weights`.
        - Cash-position for each time-step, which can be negative if
          `limit_pos > 1.0` because you are investing for borrowed money.
          If `weights` is Pandas DataFrame, then this is a Pandas Series.
          If 'weights' is Numpy 2-dim array, then this is a 1-dim array.
          If `weights` is a Pandas Series or Numpy 1-dim array,
          then this is a float value.
    """
    # Convert input weights to Numpy 2-dim array.
    weights_np = _convert_input_weights(weights)

    # Run the Jit-optimized function to normalize the weights.
    weights_norm, cash = _normalize_weights_jit(weights=weights_np,
                                                limit_pos=limit_pos,
                                                limit_neg=limit_neg,
                                                max_ratio=max_ratio)

    # Convert output weights to same data-type as input weights.
    weights_norm, cash = _convert_output_weights(weights=weights, cash=cash,
                                                 weights_norm=weights_norm)

    return weights_norm, cash


###############################################################################



# -*- coding: utf-8 -*-
from setuptools import setup

# Also defined in investops/__init__.py and must be updated in both places.
MY_VERSION = '0.2.0'

setup(
    name='investops',
    packages=['investops'],
    version=MY_VERSION,
    description='Tools for investing in Python',
    long_description='InvestOps is a collection of Python tools for investing, '
                     'including the fast and robust \"Hvass Diversification\" '
                     'algorithm for diversifying an investment portfolio.',
    long_description_content_type="text/markdown",
    author='Magnus Erik Hvass Pedersen',
    author_email='my-first-name@hvass-labs.org',
    url='https://github.com/Hvass-Labs/InvestOps',
    license='MIT',
    keywords=['investing', 'portfolio optimization', 'hvass diversification'],
    install_requires=[
        'numpy',
        'pandas',
        'numba',
    ],
)




###############################################################################
#
# Functions for sparse matrices.
#
###############################################################################
#
# This file is part of InvestOps:
#
# https://github.com/Hvass-Labs/InvestOps
#
# Published under the MIT License. See the file LICENSE for details.
#
# Copyright 2021 by Magnus Erik Hvass Pedersen
#
###############################################################################

import pandas as pd
import numpy as np
from numba import jit

###############################################################################
# Convert sparse correlation matrix.

def sparse_corr_to_numpy(weights, corr_i, corr_j, corr_coef):
    """
    Convert portfolio weights and a sparse correlation matrix to Numpy arrays.
    This is ONLY to be used when `weights` is a Pandas DataFrame or Series,
    which uses strings instead of numbers for indexing assets in the portfolio,
    e.g. when using stock-tickers such as 'MSFT' or 'AMZN'.

    The sparse matrix is in Coordinate (COO) format, where the rows are given
    in `corr_i`, and the columns are given in `corr_j`, both of which must be
    lists of strings that are found in the index of `weights`. The correlation
    coefficients for those positions in the matrix are given in `corr_coef`.

    The outputs are Numpy arrays where all the strings that identify the assets
    in the portfolio have been converted to integers that index the weights.

    :param weights:
        Pandas DataFrame or Series that maps strings such as stock-tickers
        to floating-point values for the portfolio weights.

    :param corr_i:
        Iterator with row-indices for the correlation matrix. These must
        be some of the same strings used in the index of `weights`.

    :param corr_j:
        Iterator with column-indices for the correlation matrix. These must
        be some of the same strings used in the index of `weights`.

    :param corr_coef:
        Iterator with the correlation coefficients at position (corr_i, corr_j)
        of the correlation matrix.

    :raises:
        - TypeError if `weights` is not a Pandas DataFrame or Series.
        - KeyError if a string in `corr_i` or `corr_j` is not found in
          the index of `weights`.

    :return:
        - weights_np: Numpy array with floats for the portfolio weights.
        - corr_i_np: Numpy array with ints for row-indices of corr. matrix.
        - corr_j_np: Numpy array with ints for col-indices of corr. matrix.
        - corr_coef_np: Numpy array with floats for corr. coefficients.
    """
    # Check the weights are a Pandas data-types.
    if not isinstance(weights, (pd.DataFrame, pd.Series)):
        msg = 'Argument \'weights\' must be a Pandas DataFrame or Series.'
        raise TypeError(msg)

    # Get the list of strings e.g. stock-tickers used as index in the weights.
    index = weights.index

    # Create a map from the strings used in the index to their positions.
    # This allows fast conversion of the sparse correlation matrix below.
    index_map = dict(zip(index, range(len(index))))

    # Convert weights from Pandas to Numpy. This may be a copy of the data.
    # Note: Flatten is necessary if it is a Pandas DataFrame.
    weights_np = weights.to_numpy().flatten()

    # Convert coordinates of the sparse correlation matrix to integers.
    # This raises an exception if a value is not in the index of the weights.
    corr_i_list = [index_map[i] for i in corr_i]
    corr_j_list = [index_map[j] for j in corr_j]

    # Convert to Numpy arrays.
    corr_i_np = np.asarray(corr_i_list)
    corr_j_np = np.asarray(corr_j_list)
    corr_coef_np = np.asarray(corr_coef)

    return weights_np, corr_i_np, corr_j_np, corr_coef_np

###############################################################################
# Convert between dense and sparse matrix formats.

@jit
def matrix_to_sparse(mat):
    """
    Convert a 2-dim Numpy array to a sparse matrix in Coordinate (COO) format,
    which is given by 3 arrays for the row and column indices, and the values.

    :param mat:
        Numpy 2-dim array / matrix.
    
    :return:
        - Numpy array of the row-indices i for the sparse matrix.
        - Numpy array of the column-indices j for the sparse matrix.
        - Numpy array of the non-zero values for the sparse matrix. 
    """
    # Initialize empty lists for the sparse matrix coordinates and values.
    sparse_i = []
    sparse_j = []
    sparse_v = []

    # For each row in the matrix.
    for i in range(mat.shape[0]):
        # For each column in the matrix.
        for j in range(mat.shape[1]):
            # Matrix value at this position.
            v = mat[i, j]

            # If value is non-zero then append to sparse matrix.
            if v != 0:
                sparse_i.append(i)
                sparse_j.append(j)
                sparse_v.append(v)

    # Convert to Numpy arrays.
    sparse_i = np.array(sparse_i)
    sparse_j = np.array(sparse_j)
    sparse_v = np.array(sparse_v)

    return sparse_i, sparse_j, sparse_v


@jit
def sparse_to_matrix(sparse_i, sparse_j, sparse_v, shape, dtype=np.float64):
    """
    Convert a sparse matrix in Coordinate (COO) format to a 2-dim Numpy array.

    :param sparse_i:
        Numpy array with integers for the row-indices of the sparse matrix.

    :param sparse_j:
        Numpy array with integers for the column-indices of the sparse matrix.

    :param sparse_v:
        Numpy array with floats for the values of the sparse matrix.

    :param shape:
        Tuple with the number of rows and columns for the output Numpy array.

    :param dtype:
        Data-type for the output Numpy array.

    :return:
        Numpy 2-dim array.
    """
    # Initialize the new array with zeros.
    mat = np.zeros(shape=shape, dtype=dtype)

    # Populate the Numpy array with the values from the sparse matrix.
    for i, j, v in zip(sparse_i, sparse_j, sparse_v):
        mat[i, j] = v

    return mat

###############################################################################
